import sys

import logmuse

from bedboss.cli import app

_LOGGER = logmuse.init_logger("bedboss")


def main():
    app(prog_name="bedboss")


if __name__ == "__main__":
    try:
        main()

    except KeyboardInterrupt:
        print("Pipeline aborted.")
        sys.exit(1)



from bedboss.bbuploader.main import upload_all, upload_gse


def test_manual():
    upload_gse(
        # gse="gse246900",
        # gse="gse247593",
        # gse="gse241222",
        # gse="gse266130",
        # gse="gse99178",
        gse="gse269114",
        # gse="gse240325", # TODO: check if qc works
        # gse="gse229592", # mice
        bedbase_config="/home/bnt4me/virginia/repos/bbuploader/config_db_local.yaml",
        outfolder="/home/bnt4me/virginia/repos/bbuploader/data",
        # genome="HG38",
        # rerun=True,
        standardize_pep=True,
        run_failed=True,
        run_skipped=True,
    )
    # upload_all(
    #     bedbase_config="/home/bnt4me/virginia/repos/bbuploader/config_db_local.yaml",
    #     outfolder="/home/bnt4me/virginia/repos/bbuploader/data",
    #     start_date="2024/01/21",
    #     end_date="2024/08/28",
    #     search_limit=2,
    #     search_offset=0,
    #     genome="GRCh38",
    #     rerun=True,
    # )


# upload_all(
#     bedbase_config="/home/bnt4me/virginia/repos/bbuploader/config_db_local.yaml",
#     outfolder="/home/bnt4me/virginia/repos/bbuploader/data",
#     start_date="2024/01/01",
#     # end_date="2024/03/28",
#     search_limit=200,
#     search_offset=0,
#     genome="GRCh38",
# )



from typing import Optional

from pydantic import BaseModel, ConfigDict, Field, field_validator


class BedBossMetadata(BaseModel):
    genome: str = Field(None, alias="ref_genome")
    species_name: Optional[str] = Field("", alias="sample_organism_ch1")
    species_id: Optional[str] = Field("", alias="sample_taxid_ch1")
    cell_type: Optional[str] = ""
    cell_line: Optional[str] = ""
    genotype: Optional[str] = ""
    assay: Optional[str] = Field("", alias="sample_library_strategy")
    library_source: Optional[str] = Field("", alias="sample_library_source")
    target: Optional[str] = Field("")
    antibody: Optional[str] = Field("", alias="chip_antibody")
    treatment: Optional[str] = Field("", alias="sample_treatment_protocol_ch1")
    tissue: Optional[str] = ""

    global_sample_id: Optional[str] = Field("", alias="sample_geo_accession")
    global_experiment_id: Optional[str] = Field("", alias="gse")
    description: Optional[str] = Field("", alias="sample_description")

    model_config = ConfigDict(
        populate_by_name=True,
        extra="allow",
    )

    @field_validator("global_sample_id", "global_experiment_id")
    def value_must_not_be_empty(cls, v):
        value = v.lower()
        if value.startswith("gsm") or value.startswith("gse"):
            return f"geo:{value}"
        return value


class BedBossRequired(BaseModel):
    sample_name: str
    file_path: str
    ref_genome: str
    type: Optional[str] = "bed"
    narrowpeak: Optional[bool] = False
    description: Optional[str] = ""
    organism: Optional[str] = None
    pep: Optional[BedBossMetadata] = None
    title: Optional[str] = None


class ProjectProcessingStatus(BaseModel):
    number_of_samples: int = 0
    number_of_processed: int = 0
    number_of_skipped: int = 0
    number_of_failed: int = 0



import os

from bedboss.refgenome_validator.main import ReferenceValidator

FILE_DIR = os.path.dirname(os.path.realpath(__file__))
HG19_CORRECT_DIR = os.path.join(FILE_DIR, "test_data", "bed", "hg19", "correct")
FILE_PATH = f"{HG19_CORRECT_DIR}/sample1.bed.gz"


def test_main():
    # dict_result = ReferenceValidator().determine_compatibility(
    #     FILE_PATH,
    #     concise=True,
    # )
    dict_result = ReferenceValidator().determine_compatibility(
        "/home/bnt4me/.bbcache/bedfiles/0/7/0740332b148a613342603e2e483f53e5.bed.gz",
        concise=True,
    )
    # result = ReferenceValidator().predict(
    #     # "/home/bnt4me/.bbcache/bedfiles/0/4/04c46b96264ef40bca93f03b10345da5.bed.gz",
    #     "/home/bnt4me/.bbcache/bedfiles/1/9/19ed879232e44812b1ee35b57792b924.bed.gz",
    # )
    # result

    assert dict_result



BED_TO_BIGBED_PROGRAM = "bedToBigBed"
# BED_TO_BIGBED_PROGRAM = "/home/bnt4me/virginia/repos/bedbase_all/bedboss/bedToBigBed"
BIGBED_TO_BED_PROGRAM = "bigBedToBed"


# COMMANDS TEMPLATES
# bedGraph to bed
BEDGRAPH_TEMPLATE = "macs2 {width} -i {input} -o {output}"
# bigBed to bed
BIGBED_TEMPLATE = f"{BIGBED_TO_BED_PROGRAM} {{input}} {{output}}"
# bigWig to bed
BIGWIG_TEMPLATE = (
    "bigWigToBedGraph {input} /dev/stdout | macs2 {width} -i /dev/stdin -o {output}"
)

WIG_TEMPLATE = "wigToBigWig {input} {chrom_sizes} {intermediate_bw} -clip"
# bed default link
# bed_template = "ln -s {input} {output}"
BED_TEMPLATE = "cp {input} {output}"
# gzip output files
GZIP_TEMPLATE = "gzip {unzipped_converted_file} "

# CONSTANTS
# Creating list of standard chromosome names:
STANDARD_CHROM_LIST = ["chr" + str(chr_nb) for chr_nb in list(range(1, 23))]
STANDARD_CHROM_LIST[len(STANDARD_CHROM_LIST) :] = ["chrX", "chrY", "chrM"]

QC_FOLDER_NAME = "bed_qc"

BIGBED_FOLDER_NAME = "bigbed_files"



GENOME_FILES = {
    "ensembl_hg38.chrom.sizes": "hg38",
    "ncbi_hg38.chrom.sizes": "hg38",
    "ucsc_dm6.chrom.sizes": "dm6",
    "ucsc_hg19.chrom.sizes": "hg19",
    "ucsc_hg38.chrom.sizes": "hg38",
    "ucsc_mm10.chrom.sizes": "mm10",
    "ucsc_mm39.chrom.sizes": "mm39",
    "ucsc_mm9.chrom.sizes": "mm9",
    "ucsc_panTo6.chrom.sizes": "panTro6",
}



import json
import logging
import os
import statistics
from pathlib import Path
from typing import Union

import pypiper
from geniml.io import RegionSet

from bedboss.bedstat.gc_content import calculate_gc_content, create_gc_plot
from bedboss.const import (
    BEDSTAT_OUTPUT,
    HOME_PATH,
    OPEN_SIGNAL_FOLDER_NAME,
    OPEN_SIGNAL_URL,
    OS_HG19,
    OS_HG38,
    OS_MM10,
    OUTPUT_FOLDER_NAME,
)
from bedboss.exceptions import BedBossException, OpenSignalMatrixException
from bedboss.utils import download_file
from bedboss.bedstat.r_service import RServiceManager

_LOGGER = logging.getLogger("bedboss")

SCHEMA_PATH_BEDSTAT = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "pep_schema.yaml"
)


def get_osm_path(genome: str, out_path: str = None) -> Union[str, None]:
    """
    By providing genome name download Open Signal Matrix

    :param genome: genome assembly
    :param out_path: working directory, where osm should be saved. If None, current working directory will be used
    :return: path to the Open Signal Matrix
    """
    # TODO: add more osm
    _LOGGER.info("Getting Open Signal Matrix file path...")
    if genome == "hg19" or genome == "GRCh37":
        osm_name = OS_HG19
    elif genome == "hg38" or genome == "GRCh38":
        osm_name = OS_HG38
    elif genome == "mm10" or genome == "GRCm38":
        osm_name = OS_MM10
    else:
        raise OpenSignalMatrixException(
            "For this genome open Signal Matrix was not found."
        )
    if not out_path:
        osm_folder = os.path.join(HOME_PATH, OPEN_SIGNAL_FOLDER_NAME)
    else:
        osm_folder = os.path.join(out_path, OPEN_SIGNAL_FOLDER_NAME)

    osm_path = os.path.join(osm_folder, osm_name)
    if not os.path.exists(osm_path):
        os.makedirs(osm_folder, exist_ok=True)
        download_file(
            url=f"{OPEN_SIGNAL_URL}{osm_name}",
            path=osm_path,
            no_fail=True,
        )
    _LOGGER.info(f"Open Signal Matrix file path: {osm_path}")
    return osm_path


def bedstat(
    bedfile: str,
    genome: str,
    outfolder: str,
    bed_digest: str = None,
    ensdb: str = None,
    open_signal_matrix: str = None,
    just_db_commit: bool = False,
    rfg_config: Union[str, Path] = None,
    pm: pypiper.PipelineManager = None,
    r_service: RServiceManager = None,
) -> dict:
    """
    Run bedstat pipeline - pipeline for obtaining statistics about bed files
        and inserting them into the database

    :param str bedfile: the full path to the bed file to process
    :param str bed_digest: the digest of the bed file. Defaults to None.
    :param str open_signal_matrix: a full path to the openSignalMatrix
        required for the tissue specificity plots
    :param str outfolder: The folder for storing the pipeline results.
    :param str genome: genome assembly of the sample
    :param bool just_db_commit: if True, the pipeline will only commit to the database
    :param str rfg_config: path to the refgenie config file
    :param str ensdb: a full path to the ensdb gtf file required for genomes
        not in GDdata
    :param pm: pypiper object
    :param r_service: RServiceManager object

    :return: dict with statistics and plots metadata
    """
    outfolder_stats = os.path.join(outfolder, OUTPUT_FOLDER_NAME, BEDSTAT_OUTPUT)
    try:
        os.makedirs(outfolder_stats)
    except FileExistsError:
        pass

    # TODO: osm commented to speed up code
    # find/download open signal matrix
    if not open_signal_matrix or not os.path.exists(open_signal_matrix):
        try:
            open_signal_matrix = get_osm_path(genome)
        except OpenSignalMatrixException:
            _LOGGER.warning(
                f"Open Signal Matrix was not found for {genome}. Skipping..."
            )
    # open_signal_matrix = None

    # Used to stop pipeline bedstat is used independently
    if not pm:
        stop_pipeline = True
    else:
        stop_pipeline = False

    if not bed_digest:
        bed_digest = RegionSet(bedfile).identifier

    outfolder_stats_results = os.path.abspath(os.path.join(outfolder_stats, bed_digest))
    try:
        os.makedirs(outfolder_stats_results)
    except FileExistsError:
        pass
    json_file_path = os.path.abspath(
        os.path.join(outfolder_stats_results, bed_digest + ".json")
    )
    json_plots_file_path = os.path.abspath(
        os.path.join(outfolder_stats_results, bed_digest + "_plots.json")
    )
    if not just_db_commit:
        if not pm:
            pm_out_path = os.path.abspath(
                os.path.join(outfolder_stats, "pypiper", bed_digest)
            )
            try:
                os.makedirs(pm_out_path)
            except FileExistsError:
                pass
            pm = pypiper.PipelineManager(
                name="bedstat-pipeline",
                outfolder=pm_out_path,
                pipestat_sample_name=bed_digest,
            )

        rscript_path = os.path.join(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
            "bedstat",
            "tools",
            "regionstat_cli.R",
        )
        assert os.path.exists(rscript_path), FileNotFoundError(
            f"'{rscript_path}' script not found"
        )

        if not r_service:
            try:
                _LOGGER.info("#=>>> Running local R instance!")
                command = (
                    f"Rscript {rscript_path} --bedfilePath={bedfile} "
                    f"--openSignalMatrix={open_signal_matrix} "
                    f"--outputFolder={outfolder_stats_results} --genome={genome} "
                    f"--ensdb={ensdb} --digest={bed_digest}"
                )
                pm.run(cmd=command, target=json_file_path)
            except Exception as e:
                _LOGGER.error(f"Pipeline failed: {e}")
                raise BedBossException(f"Pipeline failed: {e}")
        else:
            _LOGGER.info("#=>>> Running R service ")
            r_service.run_file(
                file_path=bedfile,
                digest=bed_digest,
                outpath=outfolder_stats_results,
                genome=genome,
                openSignalMatrix=open_signal_matrix,
                gtffile=ensdb,
            )

    data = {}
    if os.path.exists(json_file_path):
        with open(json_file_path, "r", encoding="utf-8") as f:
            data = json.loads(f.read())
    if os.path.exists(json_plots_file_path):
        with open(json_plots_file_path, "r", encoding="utf-8") as f_plots:
            plots = json.loads(f_plots.read())
    else:
        plots = []

    # unlist the data, since the output of regionstat.R is a dict of lists of
    # length 1 and force keys to lower to correspond with the
    # postgres column identifiers
    data = {k.lower(): v[0] if isinstance(v, list) else v for k, v in data.items()}
    try:
        gc_contents = calculate_gc_content(
            bedfile=bedfile, genome=genome, rfg_config=rfg_config
        )
    except BaseException as e:
        gc_contents = None

    if gc_contents:
        gc_mean = statistics.mean(gc_contents)

        data["gc_content"] = round(gc_mean, 2)

        gc_plot = create_gc_plot(
            bed_id=bed_digest,
            gc_contents=gc_contents,
            outfolder=os.path.join(outfolder_stats_results),
            gc_mean=gc_mean,
        )
        plots.append(gc_plot)

    for plot in plots:
        plot_id = plot["name"]
        data.update({plot_id: plot})

    if "md5sum" in data:
        del data["md5sum"]

    if "name" in data:
        del data["name"]

    if stop_pipeline:
        pm.stop_pipeline()

    return data



import pathlib
from enum import Enum
from typing import Union

import pypiper
from bbconf.models.bed_models import (
    BedClassification,
    BedFiles,
    BedPlots,
    BedStatsModel,
)
from pydantic import BaseModel, ConfigDict, Field

from bedboss.const import MAX_FILE_SIZE, MAX_REGION_NUMBER, MIN_REGION_WIDTH


class DATA_FORMAT(str, Enum):
    UNKNOWN = "unknown_data_format"
    UCSC_BED = "ucsc_bed"
    UCSC_BED_RS = "ucsc_bed_rs"
    BED_LIKE = "bed_like"
    BED_LIKE_RS = "bed_like_rs"
    ENCODE_NARROWPEAK = "encode_narrowpeak"
    ENCODE_NARROWPEAK_RS = "encode_narrowpeak_rs"
    ENCODE_BROADPEAK = "encode_broadpeak"
    ENCODE_BROADPEAK_RS = "encode_broadpeak_rs"
    ENCODE_GAPPEDPEAK = "encode_gappedpeak"
    ENCODE_GAPPEDPEAK_RS = "encode_gappedpeak_rs"
    ENCODE_RNA_ELEMENTS = "encode_rna_elements"
    ENCODE_RNA_ELEMENTS_RS = "encode_rna_elements_rs"


class BedMetadata(BaseModel):
    sample_name: str
    genome: str

    species_name: str = Field(
        default="", description="Name of species. e.g. Homo sapiens.", alias="organism"
    )
    species_id: str = ""
    genotype: str = Field("", description="Genotype of the sample")
    phenotype: str = Field("", description="Phenotype of the sample")

    cell_type: str = Field(
        "",
        description="specific kind of cell with distinct characteristics found in an organism. e.g. Neurons, Hepatocytes, Adipocytes",
    )
    cell_line: str = Field(
        "",
        description="population of cells derived from a single cell and cultured in the lab for extended use, e.g. HeLa, HepG2, k562",
    )
    tissue: str = Field("", description="Tissue type")

    library_source: str = Field(
        "", description="Library source (e.g. genomic, transcriptomic)"
    )
    assay: str = Field(
        "", description="Experimental protocol (e.g. ChIP-seq)", alias="exp_protocol"
    )
    antibody: str = Field("", description="Antibody used in the assay")
    target: str = Field("", description="Target of the assay (e.g. H3K4me3)")
    treatment: str = Field(
        "", description="Treatment of the sample (e.g. drug treatment)"
    )

    global_sample_id: str = Field(
        "", description="Global sample identifier. e.g. GSM000"
    )  # excluded in training
    global_experiment_id: str = Field(
        "", description="Global experiment identifier. e.g. GSE000"
    )  # excluded in training
    description: str = Field("", description="Description of the sample")

    model_config = ConfigDict(
        populate_by_name=True,
        extra="allow",
    )


class BedStatCLIModel(BaseModel):
    """
    CLI model for bedstat
    """

    bedfile: Union[str, pathlib.Path]
    genome: str
    outfolder: Union[str, pathlib.Path]
    bed_digest: str = None
    bigbed: Union[str, pathlib.Path] = None
    ensdb: str = None
    open_signal_matrix: str = None
    just_db_commit: bool = False
    pm: pypiper.PipelineManager = None

    model_config = ConfigDict(extra="ignore", arbitrary_types_allowed=True)


class BedQCCLIModel(BaseModel):
    """
    CLI model for bedqc
    """

    bedfile: Union[str, pathlib.Path]
    outfolder: Union[str, pathlib.Path]
    max_file_size: int = MAX_FILE_SIZE
    max_region_number: int = MAX_REGION_NUMBER
    min_region_width: int = MIN_REGION_WIDTH
    pm: pypiper.PipelineManager = None

    model_config = ConfigDict(extra="ignore", arbitrary_types_allowed=True)


class BedMakerCLIModel(BaseModel):
    """
    CLI model for bedmaker
    """

    input_file: Union[str, pathlib.Path]
    input_type: str
    output_path: Union[str, pathlib.Path]
    # output_bigbed: Union[str, pathlib.Path]
    # sample_name: str
    genome: str
    rfg_config: Union[str, pathlib.Path] = None
    chrom_sizes: str = None
    narrowpeak: bool = False
    # standardize: bool = False
    check_qc: bool = True
    pm: pypiper.PipelineManager = None

    model_config = ConfigDict(extra="ignore", arbitrary_types_allowed=True)


class StatsUpload(BedStatsModel):
    model_config = ConfigDict(extra="ignore")


class PlotsUpload(BedPlots):
    model_config = ConfigDict(extra="ignore")


class FilesUpload(BedFiles):
    model_config = ConfigDict(extra="ignore")


class BedClassificationUpload(BedClassification):
    model_config = ConfigDict(extra="ignore")


class BedSetAnnotations(BaseModel):
    """
    Annotations for a bedset
    """

    author: Union[str, None] = None
    source: Union[str, None] = None

    model_config = ConfigDict(extra="ignore")


class BedClassificationOutput(BaseModel):
    bed_compliance: str
    data_format: DATA_FORMAT
    compliant_columns: int
    non_compliant_columns: int



"""Package-level data"""

import logging

import coloredlogs
import logmuse

from bedboss._version import __version__

__package_name__ = "bedboss"

__author__ = [
    "Oleksandr Khoroshevskyi",
    "Michal Stolarczyk",
    "Ognen Duzlevski",
    "Jose Verdezoto",
    "Bingjie Xue",
    "Donald Campbell",
]
__email__ = "khorosh@virginia.edu"

__all__ = [
    "__version__",
    "__package_name__",
    "__author__",
]

_LOGGER = logmuse.init_logger("bedboss")
coloredlogs.install(
    logger=_LOGGER,
    datefmt="%H:%M:%S",
    fmt="[%(levelname)s] [%(asctime)s] [BEDBOSS] %(message)s",
)

_LOGGER_PIPESTAT = logging.getLogger("pipestat")
_LOGGER_PIPESTAT.propagate = False
coloredlogs.install(
    logger=_LOGGER_PIPESTAT,
    datefmt="%H:%M:%S",
    fmt="[%(levelname)s] [%(asctime)s] [PIPESTAT] %(message)s",
)

_LOGGER_GENIML = logging.getLogger("geniml")
_LOGGER_GENIML.propagate = False
coloredlogs.install(
    logger=_LOGGER_GENIML,
    datefmt="%H:%M:%S",
    fmt="[%(levelname)s] [%(asctime)s] [GENIML] %(message)s",
)

_LOGGER_BBCONF = logging.getLogger("bbconf")
_LOGGER_BBCONF.propagate = False
coloredlogs.install(
    logger=_LOGGER_BBCONF,
    datefmt="%H:%M:%S",
    fmt="[%(levelname)s] [%(asctime)s] [BBCONF] %(message)s",
)

_LOGGER_PHC = logging.getLogger("pephubclient")
_LOGGER_PHC.propagate = False
coloredlogs.install(
    logger=_LOGGER_PHC,
    datefmt="%H:%M:%S",
    fmt="[%(levelname)s] [%(asctime)s] [PEPHUBCLIENT] %(message)s",
)

_LOGGER_REF_CONF = logging.getLogger("refgenconf")
_LOGGER_REF_CONF.propagate = False
coloredlogs.install(
    logger=_LOGGER_PHC,
    datefmt="%H:%M:%S",
    fmt="[%(levelname)s] [%(asctime)s] [REFGENCONF] %(message)s",
)



import logging

from bbconf.bbagent import BedBaseAgent

_LOGGER = logging.getLogger("bedboss")


def add_to_qdrant(config: str) -> None:
    """
    Reindex all bed files in qdrant

    :param config: path to the config file
    """
    agent = BedBaseAgent(config=config)
    agent.bed.reindex_qdrant()



import logging
import subprocess
from typing import List, Union

from geniml.io import RegionSet
from gtars.models import RegionSet as GRegionSet

_LOGGER = logging.getLogger("bedboss")


def get_bed_chrom_info(bedfile: Union[str, RegionSet, GRegionSet]) -> dict:
    """
    Determine chrom lengths for bed file

    :param bedfile: RegionSet object or path to bed file
    returns dict: returns dictionary where keys are chrom names and values are the max end position of that chromosome.
    """
    if isinstance(bedfile, RegionSet):
        df = bedfile.to_pandas()
        max_end_for_each_chrom = df.groupby(0)[2].max().to_dict()
    elif isinstance(bedfile, GRegionSet):
        max_end_for_each_chrom = {}
        for region in bedfile:
            if region.chrom not in max_end_for_each_chrom:
                max_end_for_each_chrom[region.chrom] = region.end
            if region.end > max_end_for_each_chrom[region.chrom]:
                max_end_for_each_chrom[region.chrom] = region.end
    else:
        df = RegionSet(bedfile).to_pandas()
        max_end_for_each_chrom = df.groupby(0)[2].max().to_dict()

    return max_end_for_each_chrom


def run_igd_command(command):
    """
    Run IGD via a subprocess, this is a temp implementation until Rust IGD python bindings are finished.
    """

    result = subprocess.run(command, shell=True, capture_output=True, text=True)
    if result.returncode == 0:
        return result.stdout
    else:
        _LOGGER.error(f"Error running command: {result.stderr}")
        return None


def parse_IGD_output(output_str) -> Union[None, List[dict]]:
    """
    Parses IGD terminal output into a list of dicts
    Args:
      output_str: The output string from IGD

    Returns:
      A list of dictionaries, where each dictionary represents a record.
    """

    try:
        lines = output_str.splitlines()
        data = []
        for line in lines:
            if line.startswith("index"):
                continue  # Skip the header line
            elif line.startswith("Total"):
                break  # Stop parsing after the "Total" line
            else:
                fields = line.split()
                record = {
                    "index": int(fields[0]),
                    "number_of_regions": int(fields[1]),
                    "number_of_hits": int(fields[2]),
                    "file_name": fields[3],
                }
                data.append(record)
        return data
    except Exception:
        return None



import logging
import os
from typing import List, Union

import matplotlib.pyplot as plt
import seaborn as sns
from gdrs import GenomeAssembly, calc_gc_content
from matplotlib.ticker import MaxNLocator
from refgenconf import RefgenconfError
from yacman.exceptions import UndefinedAliasError

from bedboss.bedmaker.utils import get_rgc

_LOGGER = logging.getLogger("bedboss")

assembly_objects = {}


def get_genome_fasta_file(genome: str, rfg_config: str = None) -> str:
    """
    Get genome fasta file with Refgenie.

    :param genome: genome name
    :param rfg_config: path to refgenie config file
    """
    rgc = get_rgc(rfg_config=rfg_config)

    try:
        # get local path to the chrom.sizes asset
        fasta_file = rgc.seek(
            genome_name=genome,
            asset_name="fasta",
            tag_name="default",
            seek_key="fasta",
        )
        _LOGGER.info(f"fasta path: {fasta_file}")
    except (UndefinedAliasError, RefgenconfError):
        # if no local chrom.sizes file found, pull it first
        _LOGGER.info("Could not determine path to chrom.sizes asset, pulling")
        rgc.pull(genome=genome, asset="fasta", tag="default")
        fasta_file = rgc.seek(
            genome_name=genome,
            asset_name="fasta",
            tag_name="default",
            seek_key="fasta",
        )
    return fasta_file


def get_genome_assembly_obj(
    genome: str, rfg_config: str = None
) -> Union[GenomeAssembly, None]:
    """
    Get assembly object for a genome.

    :param genome: genome name
    :param rfg_config: path to refgenie config file

    :return: assembly object
    """
    if genome in assembly_objects:
        return assembly_objects[genome]
    else:
        try:
            assembly_objects[genome] = GenomeAssembly(
                get_genome_fasta_file(genome, rfg_config=rfg_config)
            )
            return assembly_objects[genome]
        except Exception as e:
            _LOGGER.error(f"Could not get assembly object for {genome}: {e}")
            return None


def calculate_gc_content(
    bedfile: str, genome: str, rfg_config: str = None
) -> Union[List[float], None]:
    """
    Calculate GC content for a bed file.

    :param bedfile: path to bed file
    :param genome: genome name
    :param rfg_config: path to refgenie config file

    :return: list of GC contents
    """

    assembly_obj = get_genome_assembly_obj(genome, rfg_config=rfg_config)
    if assembly_obj is None:
        return None

    gc_contents = calc_gc_content(
        bedfile,
        assembly_obj,
        ignore_unk_chroms=True,
    )
    return gc_contents


def create_gc_plot(
    bed_id: str, gc_contents: List[float], outfolder: str, gc_mean: float
) -> dict:
    """
    Create a GC content plot.

    :param bed_id: bed ID
    :param gc_contents: list of GC contents
    :param outfolder: path to output file
    :param gc_mean: mean GC content

    :return str: path to output file
    """
    plt.rcParams["font.size"] = 10
    plt.figure(figsize=(8, 8))
    sns.kdeplot(gc_contents, linewidth=0.8, color="black")
    plt.gca().xaxis.set_major_locator(MaxNLocator(nbins=5))

    plt.axvline(
        gc_mean,
        color="r",
        linestyle="--",
        linewidth=0.8,
        label=f"Mean: {gc_mean:.2f}",
    )
    sns.despine()
    plt.xlabel("GC Content")
    plt.legend()

    plt.title("GC Content Distribution")

    pdf_path = os.path.join(outfolder, f"{bed_id}_gccontent.pdf")
    png_path = os.path.join(outfolder, f"{bed_id}_gccontent.png")

    plt.savefig(pdf_path)
    plt.savefig(png_path)

    return {
        "name": "gccontent",
        "title": "GC Content Distribution",
        "thumbnail_path": png_path,
        "path": pdf_path,
    }



class BedBossException(Exception):
    """Exception, when bedboss fails."""

    def __init__(self, reason: str = ""):
        """
        Optionally provide explanation for exceptional condition.

        :param str reason: some context why error occurred while
        using BedBoss
        """
        super(BedBossException, self).__init__(reason)


class OpenSignalMatrixException(BedBossException):
    """Exception when Open Signal Matrix does not exist."""

    def __init__(self, reason: str = ""):
        """
        Optionally provide explanation for exceptional condition.

        :param str reason: some context why error occurred while
        using Open Signal Matrix
        """
        super(OpenSignalMatrixException, self).__init__(reason)


class QualityException(BedBossException):
    """Exception, when quality test of the bed file didn't pass."""

    def __init__(self, reason: str = ""):
        """
        Optionally provide explanation for exceptional condition.

        :param str reason: reason why quality control wasn't successful
        """
        self.reason = reason
        super(QualityException, self).__init__(reason)


class RequirementsException(BedBossException):
    """Exception, when requirement packages are not installed."""

    def __init__(self, reason: str = ""):
        """
        Optionally provide explanation for exceptional condition.

        :param str reason: additional info about requirements exception
        """
        super(RequirementsException, self).__init__(reason)


class BedTypeException(BedBossException):
    """Exception when Bed Type could not be determined."""

    def __init__(self, reason: str = ""):
        """
        Optionally provide explanation for exceptional condition.

        :param str reason: some context why error occurred while
        using Open Signal Matrix
        """
        super(BedTypeException, self).__init__(reason)


class ValidatorException(BedBossException):
    """Exception when there is an exception during refgenome validation"""

    def __init__(self, reason: str = ""):
        """
        Optionally provide explanation for exceptional condition.

        :param str reason: some context why error occurred
        """
        super(BedTypeException, self).__init__(reason)



import typer

from bedboss._version import __version__

app_bbuploader = typer.Typer(
    pretty_exceptions_short=False,
    pretty_exceptions_show_locals=False,
    help="Automatic BEDbase uploader for GEO data",
)


@app_bbuploader.command(
    help="Run bedboss uploading pipeline for specified genome in specified period of time."
)
def upload_all(
    bedbase_config: str = typer.Option(..., help="Path to bedbase config file"),
    outfolder: str = typer.Option(..., help="Path to output folder"),
    start_date: str = typer.Option(
        None, help="The earliest date when opep was updated [Default: 2000/01/01]"
    ),
    end_date: str = typer.Option(
        None, help="The latest date when opep was updated [Default: today's date]"
    ),
    search_limit: int = typer.Option(
        50, help="Limit of projects to be searched. [Default: 50]"
    ),
    search_offset: int = typer.Option(
        0, help="Limit of projects to be searched. [Default: 0]"
    ),
    download_limit: int = typer.Option(
        100, help="Limit of projects to be downloaded [Default: 100]"
    ),
    genome: str = typer.Option(
        None,
        help="Reference genome [Default: None] (e.g. hg38) - if None, all genomes will be processed",
    ),
    preload: bool = typer.Option(
        True, help="Download bedfile before caching it. [Default: True]"
    ),
    create_bedset: bool = typer.Option(
        True, help="Create bedset from bed files. [Default: True]"
    ),
    overwrite: bool = typer.Option(
        False, help="Overwrite existing bedfiles. [Default: False]"
    ),
    overwrite_bedset: bool = typer.Option(
        True, help="Overwrite existing bedset. [Default: False]"
    ),
    rerun: bool = typer.Option(False, help="Re-run all the samples. [Default: False]"),
    run_skipped: bool = typer.Option(
        True, help="Run skipped projects. [Default: False]"
    ),
    run_failed: bool = typer.Option(True, help="Run failed projects. [Default: False]"),
    standardize_pep: bool = typer.Option(
        False, help="Standardize pep with BEDMESS. [Default: False]"
    ),
    use_skipper: bool = typer.Option(
        False,
        help="Use skipper to skip projects if they were processed locally [Default: False]",
    ),
    reinit_skipper: bool = typer.Option(
        False, help="Reinitialize skipper. [Default: False]"
    ),
    lite: bool = typer.Option(
        False, help="Run the pipeline in lite mode. [Default: False]"
    ),
):
    from .main import upload_all as upload_all_function

    upload_all_function(
        bedbase_config=bedbase_config,
        outfolder=outfolder,
        start_date=start_date,
        end_date=end_date,
        search_limit=search_limit,
        search_offset=search_offset,
        download_limit=download_limit,
        genome=genome,
        create_bedset=create_bedset,
        preload=preload,
        rerun=rerun,
        run_skipped=run_skipped,
        run_failed=run_failed,
        standardize_pep=standardize_pep,
        use_skipper=use_skipper,
        reinit_skipper=reinit_skipper,
        overwrite=overwrite,
        overwrite_bedset=overwrite_bedset,
        lite=lite,
    )


@app_bbuploader.command(help="Run bedboss uploading pipeline for GSE.")
def upload_gse(
    bedbase_config: str = typer.Option(..., help="Path to bedbase config file"),
    outfolder: str = typer.Option(..., help="Path to output folder"),
    gse: str = typer.Option(
        ..., help="GSE number that can be found in pephub. eg. GSE123456"
    ),
    create_bedset: bool = typer.Option(
        True, help="Create bedset from bed files. [Default: True]"
    ),
    genome: str = typer.Option(
        None,
        help=" reference genome to upload to database. If None, all genomes will be processed",
    ),
    preload: bool = typer.Option(
        True, help="Download bedfile before caching it. [Default: True]"
    ),
    rerun: bool = typer.Option(True, help="Re-run all the samples. [Default: False]"),
    run_skipped: bool = typer.Option(
        True, help="Run skipped projects. [Default: False]"
    ),
    run_failed: bool = typer.Option(True, help="Run failed projects. [Default: False]"),
    overwrite: bool = typer.Option(
        False, help="Overwrite existing bedfiles. [Default: False]"
    ),
    overwrite_bedset: bool = typer.Option(
        True, help="Overwrite existing bedset. [Default: False]"
    ),
    standardize_pep: bool = typer.Option(
        False, help="Standardize pep with BEDMESS. [Default: False]"
    ),
    use_skipper: bool = typer.Option(
        False,
        help="Use local skipper to skip projects if they were processed locally [Default: False]",
    ),
    reinit_skipper: bool = typer.Option(
        False, help="Reinitialize skipper. [Default: False]"
    ),
    lite: bool = typer.Option(
        False, help="Run the pipeline in lite mode. [Default: False]"
    ),
):
    from .main import upload_gse as upload_gse_function

    upload_gse_function(
        bedbase_config=bedbase_config,
        outfolder=outfolder,
        gse=gse,
        create_bedset=create_bedset,
        genome=genome,
        preload=preload,
        rerun=rerun,
        run_skipped=run_skipped,
        run_failed=run_failed,
        standardize_pep=standardize_pep,
        use_skipper=use_skipper,
        reinit_skipper=reinit_skipper,
        overwrite=overwrite,
        overwrite_bedset=overwrite_bedset,
        lite=lite,
    )


def version_callback(value: bool):
    if value:
        typer.echo(f"Bedboss version: {__version__}")
        raise typer.Exit()


@app_bbuploader.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(
        None, "--version", "-v", callback=version_callback, help="App version"
    ),
):
    pass



import datetime
import logging
import os
import subprocess
from typing import Union

import bbconf
import pephubclient
import peppy
import pypiper
import yaml
from bbconf.bbagent import BedBaseAgent
from bbconf.const import DEFAULT_LICENSE
from bbconf.models.base_models import FileModel
from eido import validate_project
from geniml.bbclient import BBClient
from pephubclient.helpers import MessageHandler as m
from pephubclient.helpers import is_registry_path

from bedboss._version import __version__
from bedboss.bedbuncher import run_bedbuncher
from bedboss.bedmaker.bedmaker import make_all
from bedboss.bedstat.bedstat import bedstat
from bedboss.const import BEDBOSS_PEP_SCHEMA_PATH, PKG_NAME
from bedboss.exceptions import BedBossException
from bedboss.models import (
    BedClassificationUpload,
    BedSetAnnotations,
    FilesUpload,
    PlotsUpload,
    StatsUpload,
)
from bedboss.refgenome_validator.main import ReferenceValidator
from bedboss.skipper import Skipper
from bedboss.utils import calculate_time, get_genome_digest, standardize_genome_name
from bedboss.utils import standardize_pep as pep_standardizer
from bedboss.bedstat.r_service import RServiceManager

_LOGGER = logging.getLogger(PKG_NAME)


def requirements_check() -> None:
    """
    Check if all requirements are installed

    :return: None
    """
    _LOGGER.info("Checking requirements...")
    subprocess.run(
        [
            "bash",
            f"{os.path.dirname(os.path.abspath(__file__))}/scripts/requirements_test.sh",
        ]
    )


@calculate_time
def run_all(
    input_file: str,
    input_type: str,
    outfolder: str,
    genome: str,
    bedbase_config: Union[str, bbconf.BedBaseAgent],
    name: str = None,
    license_id: str = DEFAULT_LICENSE,
    rfg_config: str = None,
    narrowpeak: bool = False,
    check_qc: bool = True,
    validate_reference: bool = True,
    chrom_sizes: str = None,
    open_signal_matrix: str = None,
    ensdb: str = None,
    other_metadata: dict = None,
    just_db_commit: bool = False,
    force_overwrite: bool = False,
    update: bool = False,
    upload_qdrant: bool = False,
    upload_s3: bool = False,
    upload_pephub: bool = False,
    lite: bool = False,
    # Universes
    universe: bool = False,
    universe_method: str = None,
    universe_bedset: str = None,
    pm: pypiper.PipelineManager = None,
    r_service: RServiceManager = None,
) -> str:
    """
    Run bedboss: bedmaker -> bedqc -> bedclassifier -> bedstat -> upload to s3, qdrant, pephub, and bedbase.

    :param str input_file: Input file [required]
    :param str input_type: Input type [required] options: (bigwig|bedgraph|bed|bigbed|wig)
    :param str outfolder: Folder, where output should be saved  [required]
    :param str genome: genome_assembly of the sample. [required] options: (hg19, hg38, mm10) # TODO: add more
    :param str name: name of the sample (human-readable name, e.g. "H3K27ac in liver") [optional]
    :param Union[str, bbconf.BedBaseConf] bedbase_config: The path to the bedbase configuration file, or bbconf object.
    :param str license_id: license identifier [optional] (default: "DUO:0000042").; Find All licenses in bedbase.org
    :param str rfg_config: file path to the genome config file [optional]
    :param bool narrowpeak: whether the regions are narrow. Used to create bed file from bedgraph or bigwig
        (transcription factor implies narrow, histone mark implies broad peaks) [optional]
    :param bool check_qc: set True to run quality control during badmaking [optional] (default: True)
    :param bool validate_reference: set True to run genome reference validator
    :param str chrom_sizes: a full path to the chrom.sizes required for the bedtobigbed conversion [optional]
    :param str open_signal_matrix: a full path to the openSignalMatrix required for the tissue [optional]
    :param dict other_metadata: a dict containing all attributes from the sample
    :param str ensdb: a full path to the ensdb gtf file required for genomes not in GDdata [optional]
        (basically genomes that's not in GDdata)
    :param bool just_db_commit: whether just to commit the JSON to the database [Default: False]
    :param bool force_overwrite: force overwrite analysis [Default: False]
    :param bool update: whether to update the record in the database [Default: False] (if True, overwrites 'force_overwrite' and ignores it)
    :param bool upload_qdrant: whether to skip qdrant indexing [Default: False]
    :param bool upload_s3: whether to upload to s3
    :param bool upload_pephub: whether to push bedfiles and metadata to pephub [Default: False]
    :param bool lite: whether to run lite version of the pipeline [Default: False]

    :param bool universe: whether to add the sample as the universe [Default: False]
    :param str universe_method: method used to create the universe [Default: None]
    :param str universe_bedset: bedset identifier for the universe [Default: None]
    :param pypiper.PipelineManager pm: pypiper object
    :param RServiceManager r_service: RServiceManager object that will run R services
    :return str bed_digest: bed digest
    """
    if isinstance(bedbase_config, str):
        bbagent = BedBaseAgent(config=bedbase_config, init_ml=not lite)
    elif isinstance(bedbase_config, bbconf.BedBaseAgent):
        bbagent = bedbase_config
    else:
        raise BedBossException("Incorrect bedbase_config type. Exiting...")

    genome = standardize_genome_name(genome)

    _LOGGER.info(f"Input file = '{input_file}'")
    _LOGGER.info(f"Output folder = '{outfolder}'")
    _LOGGER.info(f"Sample genome = '{genome}'")

    if not pm:
        pm_out_folder = os.path.join(os.path.abspath(outfolder), "pipeline_manager")
        _LOGGER.info(f"Pipeline info folder = '{pm_out_folder}'")
        pm = pypiper.PipelineManager(
            name="bedboss-pipeline",
            outfolder=pm_out_folder,
            version=__version__,
            recover=True,
        )
        stop_pipeline = True
    else:
        stop_pipeline = False

    bed_metadata = make_all(
        input_file=input_file,
        input_type=input_type,
        output_path=outfolder,
        genome=genome,
        rfg_config=rfg_config,
        narrowpeak=narrowpeak,
        check_qc=check_qc,
        chrom_sizes=chrom_sizes,
        lite=lite,
        pm=pm,
    )
    if not other_metadata:
        other_metadata = {"sample_name": name}

    if not update:
        other_metadata["original_file_name"] = os.path.basename(input_file)

    if lite:
        statistics_dict = {}
        statistics_dict["number_of_regions"] = len(bed_metadata.bed_object)
    else:
        statistics_dict = bedstat(
            bedfile=bed_metadata.bed_file,
            outfolder=outfolder,
            genome=genome,
            ensdb=ensdb,
            bed_digest=bed_metadata.bed_digest,
            open_signal_matrix=open_signal_matrix,
            just_db_commit=just_db_commit,
            rfg_config=rfg_config,
            pm=pm,
            r_service=r_service,
        )

    if "mean_region_width" not in statistics_dict:
        statistics_dict["mean_region_width"] = (
            bed_metadata.bed_object.mean_region_width()
        )

    statistics_dict["bed_compliance"] = bed_metadata.bed_compliance
    statistics_dict["data_format"] = bed_metadata.data_format.value

    if bed_metadata.bigbed_file:
        genome_digest = get_genome_digest(genome)
    else:
        genome_digest = None

    stats = StatsUpload(**statistics_dict)
    plots = PlotsUpload(**statistics_dict)

    if bed_metadata.bigbed_file:
        big_bed = FileModel(
            name="bigbedfile",
            title="BigBed file",
            path=bed_metadata.bigbed_file,
            description="Path to the bigbed file",
            thumbnail_path=None,
            file_digest=None,
        )
    else:
        big_bed = None
    files = FilesUpload(
        bed_file=FileModel(
            name="bedfile",
            title="BED file",
            path=bed_metadata.bed_file,
            description="Path to the BED file",
            thumbnail_path=None,
            file_digest=bed_metadata.bed_object.file_digest,
        ),
        bigbed_file=big_bed,
    )

    classification = BedClassificationUpload(
        name=name or bed_metadata.bed_digest,
        genome_digest=genome_digest,
        genome_alias=genome,
        bed_compliance=bed_metadata.bed_compliance,
        data_format=bed_metadata.data_format.value,
        compliant_columns=bed_metadata.compliant_columns,
        non_compliant_columns=bed_metadata.non_compliant_columns,
        header=bed_metadata.bed_object.header,
    )

    if validate_reference:
        _LOGGER.info("Validating reference genome")
        ref_valid_stats = ReferenceValidator().determine_compatibility(
            bedfile=bed_metadata.bed_file, concise=True
        )
    else:
        ref_valid_stats = None

    if update:
        bbagent.bed.update(
            identifier=bed_metadata.bed_digest,
            stats=stats.model_dump(exclude_unset=True),
            metadata=other_metadata,
            plots=plots.model_dump(exclude_unset=True),
            files=files.model_dump(exclude_unset=True),
            classification=classification.model_dump(exclude_unset=True),
            ref_validation=ref_valid_stats,
            license_id=license_id,
            upload_qdrant=upload_qdrant and not lite,
            upload_pephub=upload_pephub,
            upload_s3=upload_s3,
            local_path=outfolder,
            overwrite=True,
            processed=not lite,
            nofail=True,
        )
    else:
        bbagent.bed.add(
            identifier=bed_metadata.bed_digest,
            stats=stats.model_dump(exclude_unset=True),
            metadata=other_metadata,
            plots=plots.model_dump(exclude_unset=True),
            files=files.model_dump(exclude_unset=True),
            classification=classification.model_dump(exclude_unset=True),
            ref_validation=ref_valid_stats,
            license_id=license_id,
            upload_qdrant=upload_qdrant and not lite,
            upload_pephub=upload_pephub,
            upload_s3=upload_s3,
            local_path=outfolder,
            overwrite=force_overwrite,
            processed=not lite,
            nofail=True,
        )

    if universe:
        bbagent.bed.add_universe(
            bedfile_id=bed_metadata.bed_digest,
            bedset_id=universe_bedset,
            construct_method=universe_method,
        )

    if stop_pipeline:
        pm.stop_pipeline()

    _LOGGER.info(f"All done! Bed digest: {bed_metadata.bed_digest}")
    return bed_metadata.bed_digest


@calculate_time
def insert_pep(
    bedbase_config: str,
    output_folder: str,
    pep: Union[str, peppy.Project],
    bedset_id: str = None,
    bedset_name: str = None,
    rfg_config: str = None,
    license_id: str = DEFAULT_LICENSE,
    create_bedset: bool = False,
    bedset_heavy: bool = False,
    check_qc: bool = True,
    ensdb: str = None,
    just_db_commit: bool = False,
    force_overwrite: bool = False,
    update: bool = False,
    upload_s3: bool = False,
    upload_pephub: bool = False,
    upload_qdrant: bool = False,
    no_fail: bool = False,
    standardize_pep: bool = False,
    lite: bool = False,
    rerun: bool = False,
    pm: pypiper.PipelineManager = None,
) -> None:
    """
    Run all bedboss pipelines for all samples in the pep file.
    bedmaker -> bedqc -> bedstat -> qdrant_indexing -> bedbuncher

    :param str bedbase_config: bedbase configuration file path
    :param str output_folder: output statistics folder
    :param Union[str, peppy.Project] pep: path to the pep file or pephub registry path
    :param str bedset_id: bedset identifier
    :param str bedset_name: bedset name
    :param str rfg_config: path to the genome config file (refgenie)
    :param str license_id: license identifier [optional] (default: "DUO:0000042").; Find All licenses in bedbase.org
        This license will be used for bedfiles where license is not provided in PEP file
    :param bool create_bedset: whether to create bedset
    :param bool bedset_heavy: whether to use heavy processing (add all columns to the database)
    :param bool upload_qdrant: whether to upload bedfiles to qdrant
    :param bool check_qc: whether to run quality control during badmaking
    :param str ensdb: a full path to the ensdb gtf file required for genomes not in GDdata
    :param bool just_db_commit: whether save only to the database (Without saving locally )
    :param bool force_overwrite: whether to overwrite the existing record
    :param bool update: whether to update the record in the database. This option will overwrite the force_overwrite option. [Default: False]
    :param bool upload_s3: whether to upload to s3
    :param bool upload_pephub: whether to push bedfiles and metadata to pephub (default: False)
    :param bool upload_qdrant: whether to execute qdrant indexing
    :param bool no_fail: whether to raise an error if bedset was not added to the database
    :param bool lite: whether to run lite version of the pipeline
    :param bool standardize_pep: whether to standardize the pep file before processing by using bedms. (default: False)
    :param bool rerun: whether to rerun processed samples
    :param pypiper.PipelineManager pm: pypiper object
    :return: None
    """

    failed_samples = []
    processed_ids = []
    if isinstance(pep, peppy.Project):
        pass
    elif isinstance(pep, str):
        if is_registry_path(pep):
            pep = pephubclient.PEPHubClient().load_project(pep)
        else:
            pep = peppy.Project(pep)
    else:
        raise BedBossException("Incorrect pep type. Exiting...")

    if standardize_pep:
        pep = pep_standardizer(pep)

    if not pm:
        pm_out_folder = os.path.join(os.path.abspath(output_folder), "pipeline_manager")
        _LOGGER.info(f"Pipeline info folder = '{pm_out_folder}'")
        pm = pypiper.PipelineManager(
            name="bedboss-pipeline",
            outfolder=pm_out_folder,
            version=__version__,
            recover=True,
        )
        stop_pipeline = True
    else:
        stop_pipeline = False

    bbagent = BedBaseAgent(bedbase_config)

    validate_project(pep, BEDBOSS_PEP_SCHEMA_PATH)

    bedset_annotation = BedSetAnnotations(**pep.config).model_dump()
    skipper = Skipper(output_folder, pep.name)

    if rerun:
        skipper.reinitialize()

    if not lite:
        r_service = RServiceManager()
    else:
        r_service = None

    for i, pep_sample in enumerate(pep.samples):
        is_processed = skipper.is_processed(pep_sample.sample_name)
        if is_processed:
            m.print_success(
                f"Skipping {pep_sample.sample_name} : {is_processed}. Already processed."
            )
            processed_ids.append(is_processed)
            continue

        m.print_success(f"Processing sample {i + 1}/{len(pep.samples)}")
        _LOGGER.info(f"Running bedboss pipeline for {pep_sample.sample_name}")
        if pep_sample.get("file_type"):
            if pep_sample.get("file_type").lower() == "narrowpeak":
                is_narrow_peak = True
            else:
                is_narrow_peak = False
        else:
            is_narrow_peak = False
        try:
            bed_id = run_all(
                input_file=pep_sample.input_file,
                input_type=pep_sample.input_type,
                genome=pep_sample.genome,
                name=pep_sample.sample_name,
                bedbase_config=bbagent,
                license_id=pep_sample.get("license_id") or license_id,
                narrowpeak=is_narrow_peak,
                chrom_sizes=pep_sample.get("chrom_sizes"),
                open_signal_matrix=pep_sample.get("open_signal_matrix"),
                other_metadata=pep_sample.to_dict(),
                outfolder=output_folder,
                rfg_config=rfg_config,
                check_qc=check_qc,
                ensdb=ensdb,
                just_db_commit=just_db_commit,
                force_overwrite=force_overwrite,
                update=update,
                upload_qdrant=upload_qdrant,
                upload_s3=upload_s3,
                upload_pephub=upload_pephub,
                universe=pep_sample.get("universe"),
                universe_method=pep_sample.get("universe_method"),
                universe_bedset=pep_sample.get("universe_bedset"),
                lite=lite,
                pm=pm,
                r_service=r_service,
            )

            processed_ids.append(bed_id)
            skipper.add_processed(pep_sample.sample_name, bed_id, success=True)

        except BedBossException as e:
            _LOGGER.error(f"Failed to process {pep_sample.sample_name}. See {e}")
            failed_samples.append(pep_sample.sample_name)
            skipper.add_failed(pep_sample.sample_name, f"{e}")

    if create_bedset:
        _LOGGER.info(f"Creating bedset from {pep.name}")
        run_bedbuncher(
            bedbase_config=bbagent,
            record_id=bedset_id or pep.name,
            bed_set=processed_ids,
            name=bedset_name or pep.name,
            output_folder=output_folder,
            description=pep.description,
            heavy=bedset_heavy,
            upload_pephub=upload_pephub,
            upload_s3=upload_s3,
            no_fail=no_fail,
            force_overwrite=force_overwrite,
            annotation=bedset_annotation,
            lite=lite,
        )
    else:
        _LOGGER.info(
            f"Skipping bedset creation. Create_bedset is set to {create_bedset}"
        )
    m.print_success(f"Processed samples: {processed_ids}")
    m.print_error(f"Failed samples: {failed_samples}")

    m.print_success(f"Processed samples: {processed_ids}")
    m.print_error(f"Failed samples: {failed_samples}")
    if stop_pipeline:
        pm.stop_pipeline()

    m.print_success(f"Processing of '{pep.name}' completed successfully")
    return None


@calculate_time
def reprocess_all(
    bedbase_config: Union[str, BedBaseAgent],
    output_folder: str,
    limit: int = 10,
    no_fail: bool = False,
    pm: pypiper.PipelineManager = None,
) -> None:
    """
    Run bedboss pipeline for all unprocessed beds in the bedbase

    :param bedbase_config: bedbase configuration file path
    :param output_folder: output folder of the pipeline
    :param limit: limit of the number of beds to process
    :param no_fail: whether to raise an error if bedset was not added to the database
    :param pm: pypiper object

    :return: None
    """

    if not pm:
        pm_out_folder = os.path.join(os.path.abspath(output_folder), "pipeline_manager")
        _LOGGER.info(f"Pipeline info folder = '{pm_out_folder}'")
        pm = pypiper.PipelineManager(
            name="bedboss-pipeline",
            outfolder=pm_out_folder,
            version=__version__,
            recover=True,
        )
        stop_pipeline = True
    else:
        stop_pipeline = False

    r_service = RServiceManager()

    if isinstance(bedbase_config, str):
        bbagent = BedBaseAgent(config=bedbase_config)
    elif isinstance(bedbase_config, bbconf.BedBaseAgent):
        bbagent = bedbase_config
    else:
        raise BedBossException("Incorrect bedbase_config type. Exiting...")

    unprocessed_beds = bbagent.bed.get_unprocessed(limit=limit)

    bbclient = BBClient()
    failed_samples = []
    for i, bed_annot in enumerate(unprocessed_beds.results):
        bed_file = bbclient.load_bed(bed_annot.id)

        m.print_success(
            f"\n#### Processing sample: {i + 1} / {len(unprocessed_beds.results)} ####"
        )
        try:
            run_all(
                input_file=bed_file.path,
                input_type="bed",
                outfolder=output_folder,
                genome=bed_annot.genome_alias,
                bedbase_config=bbagent,
                name=bed_annot.name,
                license_id=bed_annot.license_id,
                rfg_config=None,
                check_qc=False,
                validate_reference=True,
                chrom_sizes=None,
                open_signal_matrix=None,
                ensdb=None,
                other_metadata=None,
                just_db_commit=False,
                update=True,
                upload_qdrant=True,
                upload_s3=True,
                upload_pephub=True,
                lite=False,
                universe=False,
                universe_method=None,
                universe_bedset=None,
                pm=pm,
                r_service=r_service,
            )
        except Exception as e:
            _LOGGER.error(f"Failed to process {bed_annot.name}. See {e}")
            if no_fail:
                raise BedBossException(f"Failed to process {bed_annot.name}. See {e}")

            failed_samples.append(
                {
                    "id": bed_annot.id,
                    "error": e,
                }
            )

    if failed_samples:
        date_now = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        with open(
            os.path.join(output_folder, f"failed_samples_{date_now}.yaml"), "w"
        ) as file:
            yaml.dump(failed_samples, file)

            m.print_warning(f"Logs with failed samples are saved in {output_folder}")

    m.print_success(f"Processing completed successfully")

    print_values = dict(
        unprocessed_files=unprocessed_beds.count,
        processed_files=unprocessed_beds.limit,
        failed_files=len(failed_samples),
        success_files=unprocessed_beds.limit - len(failed_samples),
    )
    print_values["unprocessed_files_left"] = (
        print_values["unprocessed_files"] - print_values["processed_files"]
    )
    m.print_success(str(print_values))
    if stop_pipeline:
        pm.stop_pipeline()


@calculate_time
def reprocess_one(
    bedbase_config: Union[str, BedBaseAgent],
    output_folder: str,
    identifier: str,
    pm: pypiper.PipelineManager = None,
) -> None:
    """
    Run bedboss pipeline for one bed in the bedbase [Reprocess]

    :param bedbase_config: bedbase configuration file path
    :param output_folder: output folder of the pipeline
    :param identifier: bed identifier
    :param pm: pypiper object

    :return: None
    """

    if pm is None:
        pm_out_folder = os.path.join(os.path.abspath(output_folder), "pipeline_manager")
        _LOGGER.info(f"Pipeline info folder = '{pm_out_folder}'")
        pm = pypiper.PipelineManager(
            name="bedboss-pipeline",
            outfolder=pm_out_folder,
            version=__version__,
            recover=True,
        )
        stop_pipeline = True
    else:
        stop_pipeline = False

    if isinstance(bedbase_config, str):
        bbagent = BedBaseAgent(config=bedbase_config)
    elif isinstance(bedbase_config, bbconf.BedBaseAgent):
        bbagent = bedbase_config
    else:
        raise BedBossException("Incorrect bedbase_config type. Exiting...")

    bbclient = BBClient()

    bed_annot = bbagent.bed.get(identifier)
    bed_file = bbclient.load_bed(bed_annot.id)

    run_all(
        input_file=bed_file.path,
        input_type="bed",
        outfolder=output_folder,
        genome=bed_annot.genome_alias,
        bedbase_config=bbagent,
        name=bed_annot.name,
        license_id=bed_annot.license_id,
        rfg_config=None,
        check_qc=False,
        validate_reference=True,
        chrom_sizes=None,
        open_signal_matrix=None,
        ensdb=None,
        other_metadata=None,
        just_db_commit=False,
        update=True,
        upload_qdrant=True,
        upload_s3=True,
        upload_pephub=True,
        lite=False,
        universe=False,
        universe_method=None,
        universe_bedset=None,
        pm=None,
    )

    _LOGGER.info(f"Successfully processed {identifier}")

    if stop_pipeline:
        pm.stop_pipeline()


@calculate_time
def reprocess_bedset(
    bedbase_config: Union[str, BedBaseAgent],
    output_folder: str,
    identifier: str,
    no_fail: bool = True,
    heavy: bool = False,
):
    """
    Recalculate bedset from the bedbase

    :param bedbase_config: bedbase configuration file path
    :param output_folder: output folder of the pipeline
    :param identifier: bedset identifier
    :param no_fail: whether to raise an error if bedset was not added to the database
    :param heavy: whether to use heavy processing. Calculate plots for bedset

    :return: None
    """

    if isinstance(bedbase_config, str):
        bbagent = BedBaseAgent(config=bedbase_config)
    elif isinstance(bedbase_config, bbconf.BedBaseAgent):
        bbagent = bedbase_config
    else:
        raise BedBossException("Incorrect bedbase_config type. Exiting...")

    bedset_annot = bbagent.bedset.get(identifier)

    run_bedbuncher(
        bedbase_config=bbagent,
        record_id=bedset_annot.id,
        bed_set=bedset_annot.bed_ids,
        name=bedset_annot.name,
        output_folder=output_folder,
        description=bedset_annot.description,
        heavy=heavy,
        upload_pephub=False,
        upload_s3=heavy,
        no_fail=no_fail,
        force_overwrite=True,
        annotation={
            **bedset_annot.model_dump(
                exclude={
                    "bed_ids",
                }
            )
        },
        lite=False,
    )



# functions for tokenization of bed files
import logging
import os
from typing import Union

from bbconf.bbagent import BedBaseAgent
from geniml.bbclient import BBClient
from geniml.bbclient.const import DEFAULT_CACHE_FOLDER
from genimtools.tokenizers import TreeTokenizer

from bedboss.exceptions import BedBossException

_LOGGER = logging.getLogger("bedboss")


def tokenize_bed_file(
    universe: str,
    bed: str,
    cache_folder: Union[str, os.PathLike] = DEFAULT_CACHE_FOLDER,
    add_to_db: bool = False,
    config: str = None,
    overwrite: bool = False,
) -> None:
    """
    Tokenize all bed file and add to the local cache

    :param universe: universe name to which the bed file will be tokenized
    :param bed: bed file to be tokenized
    :param cache_folder: path to the cache folder
    :param add_to_db: flag to add tokenized bed file to the bedbase database [config should be provided if True]
    :param config: path to the bedbase config file
    :param overwrite: flag to overwrite the existing tokenized bed file

    :return: None
    """
    bbc = BBClient(cache_folder=cache_folder or DEFAULT_CACHE_FOLDER)

    tokenizer = TreeTokenizer(bbc.seek(universe))
    rs = bbc.load_bed(bed)

    tokens = tokenizer(rs).ids

    #     b = tokens.to_regions()  # [Region(chr1, 100, 200), ... ]
    #     f = tokens.to_bit_vector()  #

    bbc.cache_tokens(universe, bed, tokens)
    _LOGGER.info(f"Tokenized bed file '{bed}' added to the cache")

    if add_to_db:
        if not config:
            BedBossException(
                "Config file is required to add tokenized bed file to the database"
            )

        bbagent = BedBaseAgent(config=config)
        bbagent.bed.add_tokenized(
            bed_id=bed, universe_id=universe, token_vector=tokens, overwrite=overwrite
        )
        _LOGGER.info(f"Tokenized bed file '{bed}' added to the database")


def delete_tokenized(
    universe: str,
    bed: str,
    config: str = None,
) -> None:
    """
    Delete tokenized bed file from the database

    :param universe: universe name to which the bed file will be tokenized
    :param bed: bed file to be tokenized
    :param config: path to the bedbase config file

    :return: None
    """
    bba = BedBaseAgent(config=config)

    bba.bed.delete_tokenized(bed_id=bed, universe_id=universe)



PKG_NAME = "bbuploader"

FILE_FOLDER_NAME = "geo_files"

DEFAULT_GEO_TAG = "samples"


class STATUS:
    SUCCESS = "SUCCESS"
    FAIL = "FAIL"
    PARTIAL = "PARTIAL"
    PROCESSING = "PROCESSING"
    SKIPPED = "SKIPPED"
